/*   

  File manager file_state.c

  Contains routines to manage asynchronous file writing and reading by 
  multiple processes running on multiple processers.

  Notes for properly using the routines provided here:
 
    Routines that do writes call start_write() and use the filepointer
    returned to write data in any format, calling end_write() complete
    process.  The newly written data will not be available to any readers
    of the file, including the writing process until read_update() has been
    completed.

    Routines need to provide char *filename and, uninitialized, 
    FILE *fp and long states[2] for communication from start_write() to
    end_write().  The int retries is the number of times the writer will try
    to open a file before giving up.  On UNIX-NFS systems, no retries are
    required.
 
    A written file must be updated by a call to read_update() before a 
    subsequent write to prevent buffer being read while write procedes.  While 
    not updated, the most recently written buffer is not readable.  

    Multi-process mode:
    In the case of a read/write collision, start_write() will pause, then 
    clear the appropriate read flag and continue.  The reading process will 
    see the file has been overwritten at the end of the read.  

    Routines that do reads call start_read() and the use filepointer returned
    to read data in known format, calling end_read() to complete.
    Routines need to provide char *filename and int own_file, and
    uninitialized, FILE *fp, static unsigned long bit_flag and index, 
    int position and BOOL read_thru.  The input value, own_file, should be 1 if 
    the file is written to by this process and 0 if not. 

    If end_read() finds the buffer has been overwritten while the read was in 
    progress, it will return 0 and read_thru FALSE.
    Calling code may decide to re-read file.

  routines: 
      **** write ****
      FILE *start_write(char *filename, long states[2], int retries)
	  return of 0 means file cannot be opened, non-recoverable error
	  Checks states by reading the file.  In multi-process mode, Reads 
	  flags to determine if a read by another process is in progress, then
	  waits to see if flag clears, overwriting after WRITE_WAIT_LOOP tries. 
	  Sets file offset to write position.

	  start_write() calls 
	  int file_state_check(char* filename, long file_states[2], 
			       unsigned long flags[2]) 
	      return of 0 means file cannot be opened, non-recoverable error
	      Reads write and read states and read flags if file exists, 
	      otherwise, creates file and inits states and flags.
 
      int end_write(car *filename, FILE *fp, long *states, int retries)        
	  return of 0 means file could not be written or closed, nonrecoverable
	  Updates read and write states and closes file.

      int update_rstate(char *filename, int retries)
	  return of 0 means file could not be written or closed, nonrecoverable
	  Updates read state and closes file.
	  Must be called between file writes.

      **** read ****
      FILE *start_read(char *filename, int own_file, retries, 
		       unsigned long *bit_flag, int *index, *position)
	  return of 0 means file cannot be read now
	  Calculates and sets own read flag in file if multi-process.
	  Sets file offset to read position.

      int end_read(char *filename, FILE *fp, int own_file, retries,  
		   unsigned long bit_flag, iint index, position, BOOL *readthru)
	  return of 0 means file could not be closed, nonrecoverable, or that 
	  read needs to be repeated because of overwrite (readthru == FALSE) 
	  Checks and clears own read flag in file if multi-process, closes file.

  Additional notes:

    Considerable complication in the routines has resulted from the effort to
    keep one version that has been tested to run under Solaris and HPUX, and on
    DOS->PATHWORKS->UNIX.  The attempt to create a system that will guarantee to
    protect the integrity of data read from a file written to asynchronously by
    another process running on another processor, has not been completely
    successful, but has run without error in small-scale, heavy collision-
    producing situations.  These routines do not, as currently tested, protect
    files WRITTEN to by more than one processor, only those read by multiple
    processes, written to only by one.
    There are complicating factors in the UNIX-style buffered file IO process
    which we have dealt with by re-writing until file reads as we require.  This
    may appear unnecessary, but be aware that any changes to these checking 
    loops result in errors when multiple processes are reading and writing the 
    same files.  These checking loops are found in end_write(), read_update(),
    start_read() and end_read().  
    Because some systems return a NULL pointer on attempts to open a file being 
    read by another process, loops external to the checking loops, and limited 
    by the input variable, retries, allow repeated fopen() attempts.
    Changing calls from the fopen(), fread(), etc, family to the open(), read(),
    etc. family, did not alleviate any problems, but resulted in a slight
    degradation of performance.


  file format:
  
  position | use                    | size/type | file offset pointer
  -----------------------------------------------------------------------
      0    | read state             | 1/long    | 0
      1    | write state            | 1/long    | sizeof(long)
      2    | buffer 0 position      | 1/long    | file_pos[0] = 2*sizeof(long)
      3    | buffer 1 position      | 1/long    | file_pos[1] = 3*sizeof(long)
      4    | read flag for buffer 0 | FLG_LEN   | file_pos[2] = 4*sizeof(long)
	   |                        |  /long    |
  4+FLG_LEN| read flag for buffer 1 | FLG_LEN   | file_pos[3] 
	   |                        |  /long    |   = (4+FLG_LEN*sizeof(long)
  -----------------------------------------------------------------------
     12    | buffer 0               |  x        | file_pos[4] = .
     ...   |                        |           | buffer 0 position
     ...   |                        |           |
     ...   |                        |           |
  -----------------------------------------------------------------------
  12 + len | buffer 1               |  x        | buffer 1 position
     ...   |                        |           |
     ...   |                        |           |
 12+2*len-1| end of file            |           |
  -----------------------------------------------------------------------


  state table:

   write | read  | write       | read
   state | state | position ptr| position ptr
  -------------------------------------------
     0   |   x   | file_pos[0] | no read
	 |       |             |
     1   |   0   | file_pos[1] | no read
     1   |   1   | file_pos[1] | file_pos[0]
	 |       |             |
  %2 = 0 |   0   | file_pos[0] | file_pos[0]
  %2 = 0 |   1   | file_pos[0] | file_pos[1]
	 |       |             |
  %2 = 1 |   0   | file_pos[1] | file_pos[1]
  %2 = 1 |   1   | file_pos[1] | file_pos[0]


  Written 1995 by Leslie Thomas Wall Hoppensteadt
  Copyright 1995 by Michigan State University Board of Trustees
*/

#include "stdafx.h"
#include "MSetar.h"
#include "MSetarDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifndef __MWERKS__
  #include <sys/types.h>
  #include <sys/stat.h>
#else
  #include <types.h>
  #include <stat.h>
#endif

#include <fcntl.h>
#include <math.h>
#include "external.h"
#ifndef  HIBITL
#define  HIBITL 0x80000000
#endif
#ifndef  MAXLONG
#define MAXLONG ((long)(~HIBITL))
#endif

/* Added for compatibility with some systems where found in a strange */
/* place, 5/31/96, EDG. */
#ifndef SEEK_SET
#define SEEK_SET 0
#endif

int TotalFileOpenFailures = 0;  /* put in sga.h */

#define WRITE_WAIT_LOOP 5
#define FAIL_COUNT 50
#define FLG_LEN 4
#ifndef TRUE  /* declarations for borland c */
#define TRUE 1
#define FALSE 0
/*#define DEBUG_FILES2 */
FILE *fopen();  
#endif



static int write_wait = 0;
static long file_pos[5] = {  2*sizeof(long), 3*sizeof(long), 4*sizeof(long),
			       (4+FLG_LEN)*sizeof(long), 
			       (4+FLG_LEN*2)*sizeof(long)};














